//#define LOG_NDEBUG 0
#define LOG_TAG "OMXComponentDecodeHelper"

#include <iostream>
#include <sstream>
#include <string>
#include <dlfcn.h>
#include <linux/videodev2.h>
#include <log/log.h>
#include <ui/GraphicBufferAllocator.h>
#include <ui/GraphicBufferMapper.h>
#include <hardware/gralloc.h>
#ifdef HAL3_HWC_BUFFER
#include <hardware/GrallocWrapper.h>
#endif
#include <hardware/hardware.h>
#include <adec/osal/android_ext.h>
#include "Debug.h"
#include "OMXComponentDecodeHelper.h"
#include "OMX_Types.h"
#include "OMX_Component.h"
#include "misc.h"

#define OMX_CORE_PATH       "/vendor/lib/libOMX_Core.so"

#define MY_NAME             "Camera3"

#define USE_BUFFER          (1)

#define OMX_VERSION_MAJOR 1
#define OMX_VERSION_MINOR 1
#define OMX_VERSION_REVISION 2
#define OMX_VERSION_STEP 0

#define OMX_VERSION \
    ((OMX_VERSION_STEP<<24) | \
    (OMX_VERSION_REVISION<<16) | \
    (OMX_VERSION_MINOR<<8) | \
    OMX_VERSION_MAJOR)

#define OMX_INIT_STRUCTURE(a)                           \
    memset(&(a), 0, sizeof(a));                         \
    (a).nSize = sizeof(a);                              \
    (a).nVersion.s.nVersionMajor = OMX_VERSION_MAJOR;   \
    (a).nVersion.s.nVersionMinor = OMX_VERSION_MINOR;   \
    (a).nVersion.s.nRevision = OMX_VERSION_REVISION;    \
    (a).nVersion.s.nStep = OMX_VERSION_STEP

# define OMX_VERIFY(x) {                                                \
        const OMX_ERRORTYPE result = x;                                 \
        if(result != OMX_ErrorNone)                                     \
            ALOGE("[%s:%d]: Failure: %s (0x%x)",                        \
                __FUNCTION__,                                           \
                __LINE__,                                               \
                #x,                                                     \
                result); \
        else { OMX_CALLED(x); } }

# define OMX_CALLED(x) ALOGV("[%s:%d]: Called: %s", __FUNCTION__, __LINE__, #x)

using namespace android;

namespace androidcamera3 {

const std::string bufferReqName = MY_NAME;

const char * OMXComponentDecodeHelper::OMXEvent::describe() const
{
    std::ostringstream result;
    switch(event) {
    case OMX_EventCmdComplete:
        switch(data1) {
        case OMX_CommandStateSet:
            switch ((int)data2) {
            case OMX_StateInvalid:
                result << "StateSet:OMX_StateInvalid";
                break;
            case OMX_StateLoaded:
                result << "StateSet:OMX_StateLoaded";
                break;
            case OMX_StateIdle:
                result << "StateSet:OMX_StateIdle";
                break;
            case OMX_StateExecuting:
                result << "StateSet:OMX_StateExecuting";
                break;
            case OMX_StatePause:
                result << "StateSet:OMX_StatePause";
                break;
            case OMX_StateWaitForResources:
                result << "StateSet:OMX_StateWaitForResources";
                break;
            default:
                result << "StateSet:OMX_Statexxxxx, state(" << (int)data2 << ")";
                break;
            }
            break;
        case OMX_CommandFlush:
            result << "OMX_CommandFlushed, port(" << (int)data2 << ")";
            break;
        case OMX_CommandPortDisable:
            result << "OMX_CommandPortDisabled, nData1(0x"
                << std::hex << data1 << std::dec
                << "), port(" << (int)data2 << ")";
            break;
        case OMX_CommandPortEnable:
            result << "OMX_CommandPortEnabled, nData1(0x"
                << std::hex << data1 << std::dec << "), port("
                << (int)data2 << ")";
            break;
        case OMX_CommandMarkBuffer:
            result << "OMX_CommandMarkBuffer, nData1(0x"
                << std::hex << data1 << std::dec << "), port("
                << (int)data2 << ")";
            break;
        }
        break;
    case OMX_EventBufferFlag:
        result << "OMX_EventBufferFlag(input)";
        break;
    case OMX_EventPortSettingsChanged:
        result << "OMX_EventPortSettingsChanged(output)";
        break;
    case OMX_EventMark:
        result << "OMX_EventMarked";
        break;
    case OMX_EventResourcesAcquired:
        result << "MX_EventResourcesAcquired";
        break;
    case OMX_EventError:
        switch((OMX_S32)data1) {
        case OMX_ErrorSameState:
            break;
        case OMX_ErrorInsufficientResources:
            result << "OMX_ErrorInsufficientResources, insufficient resources";
            break;
        case OMX_ErrorFormatNotDetected:
            result << "OMX_ErrorFormatNotDetected, cannot parse input stream";
            break;
        case OMX_ErrorPortUnpopulated:
            result << "OMX_ErrorPortUnpopulated port "
                << data2 << ", cannot parse input stream";
            break;
        case OMX_ErrorStreamCorrupt:
            result << "OMX_ErrorStreamCorrupt, Bitstream corrupt";
            break;
        default:
            result << "OMX_EventError detected, nData1(0x"
                << std::hex << data1 << std::dec
                << "), port(" << (int)data2 << ")";
            break;
        }
        break;
    default:
        result << "Unknown, nData1(0x"
            << std::hex << data1 << std::dec
            << "), port(" << (int)data2 << ")";
        break;
    }
    return result.str().c_str();
}

OMX_ERRORTYPE
OMXComponentDecodeHelper::omxil_EventHandler(
    OMX_HANDLETYPE,
    OMX_PTR pAppData,
    OMX_EVENTTYPE eEvent,
    OMX_U32 nData1,
    OMX_U32 nData2,
    OMX_PTR pEventData)
{

    ALOGV("%s called",__FUNCTION__);

    OMXComponentDecodeHelper *component = (OMXComponentDecodeHelper*)pAppData;
    const OMXEvent event(eEvent, nData1, nData2, pEventData);
    return component->eventHandler(event);
}

OMXComponentDecodeHelper::OMXComponentDecodeHelper (
    OMXDecodeHelperCallback *callback,
    int32_t format,
    int w,
    int h)
: omx_init(NULL),
  omx_deinit(NULL),
  omx_get_handle(NULL),
  omx_free_handle(NULL),
  mDlHandle(NULL),
  mHandle(0),
  mInput(0),
  mOutput(0),
  mExecuting(false),
  in_port_def(),
  out_port_def(),
  mOMXCallbacks()
{
    ALOGD("OMXComponentDecodeHelper()");

    mFormat = format;
    mWidth = w;
    mHeight = h;
    mCallback = callback;
#ifdef HAL3_HWC_BUFFER
    pGrallocWrapper = GrallocWrapper_new();
#endif
    if(mFormat == V4L2_PIX_FMT_H264) {
        mCodecType = OMX_VIDEO_CodingAVC;
        mComponentName = "OMX.realtek.video.dec.avc";
    }else if (mFormat == V4L2_PIX_FMT_MJPEG) {
        mCodecType = OMX_VIDEO_CodingMJPEG;
#ifdef HAL3_HWC_BUFFER
        mComponentName = "OMX.realtek.video.dec.mjpg";
#else
        mComponentName = "OMX.realtek.video.decoder";
#endif
    }else{
        CHECK(false);
    }

    for(int i=0;i<ALLOCATOR_BUF_NUM;i++) {
        GrallocBufInfo *pt = &(mGrallocBufs[i]);
        memset(pt,0x0,sizeof(GrallocBufInfo));
    }

    init();
}

OMXComponentDecodeHelper::~OMXComponentDecodeHelper()
{
    ALOGD("~OMXComponentDecodeHelper()");
    release();

#if (USE_BUFFER)
    releaseGrallocBuf();
#endif
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper_delete(pGrallocWrapper);
#else
    //delete pGraphicBuffer;
#endif
}

status_t
OMXComponentDecodeHelper::init()
{
    // load component so
#ifdef HAL3_HWC_BUFFER
    mDlHandle = dlopen (OMX_CORE_PATH, RTLD_LAZY);
    ALOGV("init - mDlHandle %p",mDlHandle);

    omx_init = (OMX_ERRORTYPE (*)())
	    dlsym (mDlHandle, "OMX_Init");
    omx_deinit = (OMX_ERRORTYPE (*)())
    	dlsym (mDlHandle, "OMX_Deinit");
    omx_get_handle = (OMX_ERRORTYPE (*)(OMX_HANDLETYPE* , OMX_STRING, OMX_PTR, OMX_CALLBACKTYPE*))
    	dlsym (mDlHandle, "OMX_GetHandle");
    omx_free_handle = (OMX_ERRORTYPE (*)(OMX_HANDLETYPE))
    	dlsym (mDlHandle, "OMX_FreeHandle");

#else
    mDlHandle = dlopen(OMX_CORE_PATH, RTLD_NOW);
    ALOGD("init - mDlHandle %p",mDlHandle);

    omx_init = (OMX_ERRORTYPE (*)())
         dlsym (mDlHandle, "RTKOMX_Init");
    omx_deinit = (OMX_ERRORTYPE (*)())
    	dlsym (mDlHandle, "RTKOMX_DeInit");
    omx_get_handle = (OMX_ERRORTYPE (*)(OMX_HANDLETYPE* , OMX_STRING, OMX_PTR, OMX_CALLBACKTYPE*))
    	dlsym (mDlHandle, "RTKOMX_GetHandle");
    omx_free_handle = (OMX_ERRORTYPE (*)(OMX_HANDLETYPE))
    	dlsym (mDlHandle, "RTKOMX_FreeHandle");
#endif


    CHECK(omx_init != NULL);
    CHECK(omx_deinit != NULL);
    CHECK(omx_get_handle != NULL);
    CHECK(omx_free_handle != NULL);

    omx_init();

    memset(&mOMXCallbacks, 0, sizeof(mOMXCallbacks));
    mOMXCallbacks.EventHandler = OMXComponentDecodeHelper::omxil_EventHandler;
    mOMXCallbacks.EmptyBufferDone = OMXComponentDecodeHelper::omxil_EmptyBufferDone;
    mOMXCallbacks.FillBufferDone = OMXComponentDecodeHelper::omxil_FillBufferDone;

    omx_get_handle(
        &mHandle,
        mComponentName,
        this,
        &mOMXCallbacks);

    OMX_INDEXTYPE index = OMX_IndexParamVideoInit;

    OMX_PORT_PARAM_TYPE port;
    OMX_INIT_STRUCTURE(port);
    OMX_VERIFY(OMX_GetParameter(mHandle, index, &port));

    mInput  = port.nStartPortNumber;
    mOutput = mInput + 1;

    if(mOutput > port.nStartPortNumber+port.nPorts-1){
        mOutput = port.nStartPortNumber+port.nPorts-1;
    }

    char compName[128];
    OMX_UUIDTYPE uid;
    OMX_VERSIONTYPE compVersion, specVersion;
    OMX_VERIFY(OMX_GetComponentVersion(
        mHandle,
        compName,
        &compVersion,
        &specVersion,
        &uid));

    ALOGD("[%s:%d] %s: Created: [%s](0x%x/0x%x) [%d->%d]",
        __FUNCTION__,
        __LINE__,
        MY_NAME,
        compName,
        compVersion,
        specVersion,
        mInput,
        mOutput);

    return OK;
}

status_t
OMXComponentDecodeHelper::start()
{
    ALOGV("start : E");

    CHECK(mEventThread == NULL);

    mEventThread = new EventThread(this);
    mEventThread->run("OMXComponent");

    ALOGV("enablePort - inputPort");

    bool waitNeeded = enablePort(inputPort());

    if(waitNeeded) {
        Mutex::Autolock l(mEventLock);
        waitForCommandComplete(OMX_CommandPortEnable, inputPort(), 5*1E9);
    }

    ALOGV("enablePort - outputPort");
    waitNeeded = enablePort(outputPort());

    if(waitNeeded) {
        Mutex::Autolock l(mEventLock);
        waitForCommandComplete(OMX_CommandPortEnable, outputPort(), 5*1E9);
    }

    ALOGV("setOutFormat");

    OMX_VIDEO_PARAM_PORTFORMATTYPE out_format;
    get_omx_video_param(outputPort(), &out_format);
    out_format.eColorFormat = OMX_COLOR_FormatYUV420SemiPlanar;
    update_omx_video_param(outputPort(), &out_format);

    OMX_VIDEO_CODINGTYPE codec_type = mCodecType;

    OMX_PARAM_PORTDEFINITIONTYPE in_port_def;	
    get_omx_port_definition(inputPort(), &in_port_def);
    in_port_def.format.video.eCompressionFormat = codec_type;
    update_omx_port_definition(inputPort(), &in_port_def);

    OMX_PARAM_PORTDEFINITIONTYPE out_port_def;	
    get_omx_port_definition(outputPort(), &out_port_def);
    update_omx_port_definition(outputPort(), &out_port_def);

    setState(OMX_StateIdle);

    ALOGV("initInputBuffers");
    initInputBuffers();

    // NOTE: in default, always use allocat buffer mode
    ALOGV("initOutputBuffers");
    initOutputBuffers();

    {
        // Note: make sure to wait state change complete.
        Mutex::Autolock l(mEventLock);
        waitForCommandComplete(OMX_CommandStateSet, OMX_StateIdle,5*1E9);
    }

    ALOGV("setState -> OMX_StateExecuting");
    setState(OMX_StateExecuting);

    {
        // Note: make sure to wait state change complete.
        Mutex::Autolock l(mEventLock);
        waitForCommandComplete(OMX_CommandStateSet, OMX_StateExecuting, 5*1E9);
    }

    for(int i = 0; i < out_port_def.nBufferCountActual; i++){
       OMX_BUFFERHEADERTYPE *out_buffer = outputBuffer();
       fillBuffer(out_buffer);
    }

    // start input thread
    mInputBufferThread = new InputBufferThread(this);
    mInputBufferThread->run("inputBuffer");

    ALOGV("start : X");
    mExecuting = true;
    return OK;
}

status_t
OMXComponentDecodeHelper::stop()
{
    ALOGD("stop : E");
    removeThreads();
    return OK;
}

status_t
OMXComponentDecodeHelper::release()
{
    removeThreads();

    if(mHandle) {
        for(List<Connection>::iterator it = mConnections.begin(); it != mConnections.end(); ++it) {
            disablePort(it->port);
            it->component->disablePort(it->componentPort);
        }
        mConnections.clear();
        cleanupBuffers();
        omx_free_handle(mHandle); 
        mInput = mOutput = 0;
        mHandle = 0;
    }

    CHECK(omx_deinit != NULL);
    omx_deinit();

    return OK;
}

void
OMXComponentDecodeHelper::removeThreads()
{
    if(mInputBufferThread.get() != NULL) {
        mInputBufferThread->stop();
        mInputBufferThread->requestExitAndWait();
        mInputBufferThread.clear();
        mInputBufferThread = NULL;
    }

    // TODO clear EventThread
    if(mEventThread.get() != NULL) {
        mEventThread->stop();
        mEventThread->requestExitAndWait();
        mEventThread.clear();
        mEventThread = NULL;
    }
}

bool
OMXComponentDecodeHelper::emptyBuffer(OMX_BUFFERHEADERTYPE *buffer)
{
    if(handle()) {

	    ALOGV("CAMIO OMX_EmptyThisBuffer %d datasize %d",
            mInputBuffers.size(),
            buffer->nFilledLen);

        const OMX_ERRORTYPE emptied = OMX_EmptyThisBuffer(handle(), buffer);
        CHECK(buffer->nAllocLen >= buffer->nFilledLen);

        if(emptied == OMX_ErrorNone)
            return true;

        ALOGE("[%s:%d]:%s:%p: FAILURE 0x%x",
            __FUNCTION__,
            __LINE__,
            MY_NAME,
            handle(),
            emptied);
    }
    return false;
}

bool
OMXComponentDecodeHelper::fillBuffer(OMX_BUFFERHEADERTYPE *buffer)
{
    if(handle()) {
        Mutex::Autolock l(mOutBufferLock);
	    ALOGV("[realtek] %s OMX_FillThisBuffer %d",
            MY_NAME,
            mOutputBuffers.size());

        const OMX_ERRORTYPE filled = OMX_FillThisBuffer(handle(), buffer);

        if(filled == OMX_ErrorNone)
            return true;

        ALOGE("[%s:%d]:%s:%p: FAILURE 0x%x",
            __FUNCTION__,
            __LINE__,
            MY_NAME,
            handle(),
            filled);
    }
    return false;
}

OMX_BUFFERHEADERTYPE *
OMXComponentDecodeHelper::inputBuffer()
{
    Mutex::Autolock l(mInBufferLock);
    if(!mInputBuffers.empty()) {

        OMX_BUFFERHEADERTYPE *buffer = mInputBuffers[0];
        mInputBuffers.removeAt(0);

        ALOGV("CAMIO [%s] return buffer %p",
            __FUNCTION__,
            buffer);

        return buffer;
    }
    return NULL;
}

// a loop to consume output buffer
OMX_BUFFERHEADERTYPE *
OMXComponentDecodeHelper::outputBuffer()
{
    while(true) {
        Mutex::Autolock l(mOutBufferLock);
        if(!mOutputBuffers.empty()) {
            OMX_BUFFERHEADERTYPE *buffer = mOutputBuffers[0];
            mOutputBuffers.removeAt(0);
            return buffer;
        }

        ALOGD("wait output buffer timeout");
        mOutputBufferCondition.waitRelative(mOutBufferLock,1E9);
    }
    return 0;
}

bool
OMXComponentDecodeHelper::setState(OMX_STATETYPE state)
{
    if(handle()) {
        OMX_STATETYPE state_actual = OMX_StateMax;
        OMX_VERIFY(OMX_GetState(handle(), &state_actual));

        ALOGV("setState actual %d state %d",
            state_actual,
            state);

        if(state != state_actual) {
            const OMX_ERRORTYPE stateSet = OMX_SendCommand(handle(), OMX_CommandStateSet, state, 0);
            if(stateSet == OMX_ErrorNone){
                OMX_VERIFY(OMX_GetState(handle(), &state_actual));
                ALOGV("State after set %d",state_actual);
                return true;
            }

            ALOGE("[%s:%d]:%s:%p: Failure settting state(%d/0x%x)",
                __FUNCTION__,
                __LINE__,
                MY_NAME,
                handle(),
                state,
                stateSet);
        } else {
            ALOGW("[%s:%d]:%s:%p: Already in state(%d)",
                __FUNCTION__,
                __LINE__,
                MY_NAME,
                handle(),
                state);
        }
    }
    return false;
}

void
OMXComponentDecodeHelper::get_omx_port_definition(
    int port_index,
    OMX_PARAM_PORTDEFINITIONTYPE *port_def)
{
    OMX_INIT_STRUCTURE(*port_def);
	port_def->nPortIndex = port_index;	
	OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, port_def);

    ALOGV("[realtek] %s get_deinition port: %d nPortIndex %d",
        MY_NAME,
        port_index,
        port_def->nPortIndex);

    ALOGV("[realtek] %s get_deinition port: %d nBufferCountActual %d",
        MY_NAME,
        port_index,
        port_def->nBufferCountActual);

    ALOGV("[realtek] %s get_deinition port: %d nBufferCountMin %d",
        MY_NAME,
        port_index,
        port_def->nBufferCountMin);

    ALOGV("[realtek] %s get_deinition port: %d nBufferSize %d",
        MY_NAME,
        port_index,
        port_def->nBufferSize);

    ALOGV("[realtek] %s get_deinition port: %d bEnabled %d",
        MY_NAME,
        port_index,
        port_def->bEnabled);
}

void
OMXComponentDecodeHelper::update_omx_port_definition(
    int port_index,
    OMX_PARAM_PORTDEFINITIONTYPE *port_def)
{
	OMX_SetParameter(handle(), OMX_IndexParamPortDefinition, port_def);
	OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, port_def);
	if(port_index == inputPort()){
        memcpy(&in_port_def, port_def, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
	}else if(port_index == outputPort()){
        memcpy(&out_port_def, port_def, sizeof(OMX_PARAM_PORTDEFINITIONTYPE));
    }

    ALOGV("[realtek] %s update_definition port: %d nPortIndex %d",
        MY_NAME,
        port_index,
        port_def->nPortIndex);

    ALOGV("[realtek] %s update_definition port: %d nBufferCountActual %d",
        MY_NAME,
        port_index,
        port_def->nBufferCountActual);

    ALOGV("[realtek] %s update_definition port: %d nBufferCountMin %d",
        MY_NAME,
        port_index,
        port_def->nBufferCountMin);

    ALOGV("[realtek] %s update_definition port: %d nBufferSize %d",
        MY_NAME,
        port_index,
        port_def->nBufferSize);

    ALOGV("[realtek] %s update_definition port: %d bEnabled %d",
        MY_NAME,
        port_index,
        port_def->bEnabled);
}

void
OMXComponentDecodeHelper::update_omx_video_param(
    int port_index,
    OMX_VIDEO_PARAM_PORTFORMATTYPE *format)
{
	OMX_SetParameter(handle(), OMX_IndexParamVideoPortFormat , format);
	OMX_GetParameter(handle(), OMX_IndexParamVideoPortFormat, format);
	ALOGD("[realtek] %s update_param port: %d eColorFormat %d",
        MY_NAME,
        port_index,
        format->eColorFormat);
}

void
OMXComponentDecodeHelper::get_omx_video_param(
    int port_index,
    OMX_VIDEO_PARAM_PORTFORMATTYPE *format)
{
    OMX_INIT_STRUCTURE(*format);
    format->nPortIndex = port_index;
    OMX_GetParameter(handle(), OMX_IndexParamVideoPortFormat, format);

    ALOGD("[realtek] %s get_param port: %d eColorFormat %d",
        MY_NAME,
        port_index,
        format->eColorFormat);
}

bool
OMXComponentDecodeHelper::waitForCommandComplete(
    OMX_COMMANDTYPE command,
    OMX_U32 nData,
    nsecs_t wait)
{
    bool result = false;
    {
        //Mutex::Autolock l(mEventLock);
        while(true) {
            if(!mEvents.empty()) {
                for(List<OMXEvent>::iterator it = mEvents.begin(); it != mEvents.end(); ++it) {
                    if(it->event == OMX_EventCmdComplete && it->data1 == command && it->data2 == nData) {
                        mEvents.erase(it);
                        result = true;
                        break;
                    }
                }
                if(result)
                    break;
            }
            if(mEventCondition.waitRelative(mEventLock, wait) == TIMED_OUT)
                break;
        }
    }
    return result;
}

bool
OMXComponentDecodeHelper::enablePort(uint32_t port)
{
    if(handle()) {
        OMX_PARAM_PORTDEFINITIONTYPE portFormat;
        OMX_INIT_STRUCTURE(portFormat);
        portFormat.nPortIndex = port;

        OMX_VERIFY(OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, &portFormat));

        ALOGV("[%s] portFormat.bEnabled %d",
            __FUNCTION__,
            portFormat.bEnabled);

        if(portFormat.bEnabled == OMX_FALSE) {
            OMX_VERIFY(OMX_SendCommand(handle(), OMX_CommandPortEnable, port, 0));
            return true;
        }else{
            // don't have to wait
            return false;
        }
    }else{
        CHECK(false);
        return false;
    }
}

bool
OMXComponentDecodeHelper::disablePort(uint32_t port)
{
    if(handle()) {

        OMX_VERIFY(OMX_SendCommand(
            handle(),
            OMX_CommandPortDisable,
            port,
            0));

        return true;
    }else{
        CHECK(false);
        return false;
    }
}

int
OMXComponentDecodeHelper::initInputBuffers(int port)
{
    if(port == -1)
        port = mInput;

    if(handle()) {

        OMX_PARAM_PORTDEFINITIONTYPE portFormat;
        OMX_INIT_STRUCTURE(portFormat);
        portFormat.nPortIndex = port;

        OMX_VERIFY(OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, &portFormat));

        ALOGV("initInputBuffers nBufferCountActual %d",portFormat.nBufferCountActual);

        for(unsigned int i = 0; i < portFormat.nBufferCountActual; ++i) {
            OMX_BUFFERHEADERTYPE *buffer = NULL;

            const OMX_ERRORTYPE created =
            OMX_AllocateBuffer(
                handle(),
                &buffer,
                port,
                0,
                portFormat.nBufferSize);

            if(created == OMX_ErrorNone) {
                Mutex::Autolock l(mInBufferLock);
                mInputBuffers.push_back(buffer);
                mInputBufferCondition.signal();
            } else {
                ALOGE("[%s]Failure(%d/0x%x)",__FUNCTION__, i, created);
            }
        }

        if(!mInputBuffers.empty())
            return mInputBuffers.size();
    }
    return 0;
}

void
OMXComponentDecodeHelper::allocateGrallocBuf(
    size_t reqSize,
    size_t bufNum)
{
    // release buffer first
    releaseGrallocBuf();

    CHECK(ALLOCATOR_BUF_NUM >= bufNum);

    status_t err=0;
    uint32_t stride=0;
    uint32_t phys_info[2];
    buffer_handle_t buf;

    int width = ROUNDUP16(mWidth);
    int height = ROUNDUP16(mHeight);
#ifdef HAL3_HWC_BUFFER
    int usage =
          GRALLOC_USAGE_HW_ION
        | GRALLOC_USAGE_HW_COMPOSER
        | GRALLOC_USAGE_SW_READ_OFTEN
        | GRALLOC_USAGE_SW_WRITE_OFTEN;
#else
    int usage = GRALLOC_USAGE_HW_COMPOSER
        | GRALLOC_USAGE_SW_READ_OFTEN
        | GRALLOC_USAGE_SW_WRITE_OFTEN;
#endif

    GraphicBufferAllocator &GrallocAlloc = GraphicBufferAllocator::get();
    GraphicBufferMapper &mapper = GraphicBufferMapper::get();

    for(int i=0;i<bufNum;i++) {
        uint32_t phys_info[2];
        GrallocBufInfo *pt = &(mGrallocBufs[i]);
#ifdef HAL3_HWC_BUFFER
        err = GrallocAlloc.allocate(
                width,
                height,
                HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP,
                1, // layer count
                usage,
                &(pt->buffer),
                &stride,
                i,
                bufferReqName);
#else
         ALOGD("allocateGrallocBuf 1");
         err = GrallocAlloc.allocate(
                width,
                height,
                HAL_PIXEL_FORMAT_YCrCb_420_SP,
                1, // layer count
                usage,
                &(pt->buffer),
                &stride,
                i,
                bufferReqName);
#endif
        CHECK(err == NO_ERROR);
#ifdef HAL3_HWC_BUFFER
        GrallocWrapper_GetPhyAddrs(pGrallocWrapper,pt->buffer,phys_info);
#endif
        pt->phyAddr = phys_info[0];

        Rect bounds((uint32_t)width, (uint32_t)height);
        mapper.lock(pt->buffer,
            GRALLOC_USAGE_SW_READ_OFTEN | GRALLOC_USAGE_SW_WRITE_OFTEN,
            bounds,
            &(pt->vAddr));
#ifdef HAL3_HWC_BUFFER
        size_t size = pGrallocWrapper->GetSize(pt->buffer);
        pt->size = size;
        ALOGD("[%s] buffer_handle_t %p, vaddr %p phyAddr %.8x",
            __FUNCTION__,
            pt->buffer,
            pt->vAddr,
            pt->phyAddr);
#endif
    }
}

void
OMXComponentDecodeHelper::releaseGrallocBuf()
{
    GraphicBufferAllocator &GrallocAlloc = GraphicBufferAllocator::get();
    for(int i=0;i<ALLOCATOR_BUF_NUM;i++) {
        GrallocBufInfo *pt = &(mGrallocBufs[i]);
        if(pt->buffer != NULL) {
            ALOGD("%s free buffer %p",__FUNCTION__,pt->buffer);
            GrallocAlloc.free(pt->buffer);
        }
        memset(pt,0x0,sizeof(GrallocBufInfo));
    }
}

int
OMXComponentDecodeHelper::initOutputBuffers2(int port)
{
    if(port == -1)
        port = mOutput;
    if(handle()) {
        OMX_PARAM_PORTDEFINITIONTYPE portFormat;
        OMX_INIT_STRUCTURE(portFormat);
        portFormat.nPortIndex = port;

        OMX_VERIFY(OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, &portFormat));

        ALOGV("[%s] nBufferCountActual : %d nBufferSize : %d allocateSize : %d",
            __FUNCTION__,
            portFormat.nBufferCountActual,
            portFormat.nBufferSize,
            mGrallocBufs[0].size);

        CHECK(portFormat.nBufferCountActual <= ALLOCATOR_BUF_NUM);

        allocateGrallocBuf(portFormat.nBufferSize,
            portFormat.nBufferCountActual);

        for(unsigned int i = 0; i < portFormat.nBufferCountActual; ++i) {

            CHECK(mGrallocBufs[i].buffer != NULL);
            OMX_BUFFERHEADERTYPE *buffer = NULL;

            const OMX_ERRORTYPE eError =
                OMX_UseBuffer(
                    handle(),
                    &buffer,
                    port,
                    0,
                    portFormat.nBufferSize,
                    (OMX_U8 *) mGrallocBufs[i].buffer);

            if(eError == OMX_ErrorNone) {
                Mutex::Autolock l(mOutBufferLock);
                mOutputBuffers.push_back(buffer);
            } else {
                ALOGE("[%s]Failure(%d/0x%x)",__FUNCTION__, i, eError);
            }
        }

        if(!mOutputBuffers.empty())
            return mOutputBuffers.size();
    }

    return 0;
}

int
OMXComponentDecodeHelper::initOutputBuffers(int port)
{
    if(port == -1)
        port = mOutput;
    if(handle()) {

        OMX_PARAM_PORTDEFINITIONTYPE portFormat;
        OMX_INIT_STRUCTURE(portFormat);
        portFormat.nPortIndex = port;

        OMX_VERIFY(OMX_GetParameter(handle(), OMX_IndexParamPortDefinition, &portFormat));

        for(unsigned int i = 0; i < portFormat.nBufferCountActual; ++i) {
            OMX_BUFFERHEADERTYPE *buffer = NULL;
            const OMX_ERRORTYPE created =
                OMX_AllocateBuffer(
                    handle(),
                    &buffer,
                    port,
                    0,
                    portFormat.nBufferSize);
            if(created == OMX_ErrorNone) {
                Mutex::Autolock l(mOutBufferLock);
                mOutputBuffers.push_back(buffer);
            } else {
                ALOGE("[%s]Failure(%d/0x%x)",__FUNCTION__, i, created);
            }
        }

        if(!mOutputBuffers.empty())
            return mOutputBuffers.size();
    }
    return 0;
}

void
OMXComponentDecodeHelper::cleanupOutputBuffer(int port)
{
    if(port == -1)
        port = mOutput;

    if(!mOutputBuffers.empty()) {
        Mutex::Autolock l(mOutBufferLock);
        for(int i = 0; !mOutputBuffers.empty(); ++i) {
            OMX_BUFFERHEADERTYPE *buffer = mOutputBuffers[0];
            mOutputBuffers.removeAt(0);
            const OMX_ERRORTYPE destroyed = OMX_FreeBuffer(handle(), port, buffer);
            if(destroyed == OMX_ErrorNone) {
            } else {
                ALOGE("[%s] Failure(%d/0x%x)",__FUNCTION__, i, destroyed);
            }
        }
    }
}

void
OMXComponentDecodeHelper::cleanupInputBuffers(int port)
{
    if(port == -1)
        port = mInput;

    if(!mInputBuffers.empty()) {
        Mutex::Autolock l(mInBufferLock);
        for(int i = 0; !mInputBuffers.empty(); ++i) {
            OMX_BUFFERHEADERTYPE *buffer = mInputBuffers[0];
            mInputBuffers.removeAt(0);
            const OMX_ERRORTYPE destroyed = OMX_FreeBuffer(handle(), port, buffer);
            if(destroyed == OMX_ErrorNone) {
            } else {
                ALOGE("[%s] Failure(%d/0x%x)",__FUNCTION__, i, destroyed);
            }
        }
    }
}

void
OMXComponentDecodeHelper::cleanupBuffers()
{
    cleanupOutputBuffer();
    cleanupInputBuffers();
}

OMX_ERRORTYPE
OMXComponentDecodeHelper::eventHandler(OMXEvent event)
{
    ALOGV("eventHandler %s",event.describe());

    Mutex::Autolock l(mEventLock);
    mEvents.push_back(event);
    mEventCondition.broadcast();

    return OMX_ErrorNone;
}

void
OMXComponentDecodeHelper::filledBuffer(OMX_BUFFERHEADERTYPE *buffer)
{

    ALOGV("[%s]: E buffer->pBuffer %p",
        __FUNCTION__,
        buffer->pBuffer);

    uint32_t phyAddr = 0;
    void *vAddr = NULL;

#if (USE_BUFFER)
    for(int i=0;i<ALLOCATOR_BUF_NUM;i++) {
        if(buffer->pBuffer == (OMX_U8 *) mGrallocBufs[i].buffer) {
            phyAddr = mGrallocBufs[i].phyAddr;
            vAddr = mGrallocBufs[i].vAddr;

            ALOGV("filledBuffer - find phyAddr %.8x filledLen %d pts %lld",
                phyAddr,
                buffer->nFilledLen,
                buffer->nTimeStamp);
            break;
        }
    }
#endif

    if(phyAddr != 0 && vAddr != NULL && buffer->nFilledLen > 0 && buffer->nTimeStamp > 0) {
        // report only if it is a vailded buffer
        mCallback->onOutputBufferAvailable(
            (uint8_t *) vAddr,
            phyAddr,
            buffer->nFilledLen,
            buffer->nTimeStamp);
    }

    {
        Mutex::Autolock l(mOutBufferLock);
        mOutputBuffers.push_back(buffer);
        mOutputBufferCondition.signal();
    }
    ALOGV("[%s]: X",__FUNCTION__);
}

void
OMXComponentDecodeHelper::emptiedBuffer(OMX_BUFFERHEADERTYPE *buffer)
{
    ALOGV("emptiedBuffer buffer %p",buffer);

    {
        Mutex::Autolock l(mInBufferLock);
        mInputBuffers.push_back(buffer);
        ALOGV("[%s] mInputBuffers size %d",
            __FUNCTION__,
            mInputBuffers.size());
        mInputBufferCondition.signal();
    }

    //emptyBuffer(buffer);
}

void
OMXComponentDecodeHelper::flush(int port)
{
    if(handle()) {
        OMX_VERIFY(OMX_SendCommand(handle(), OMX_CommandFlush, port, 0));
        {
            Mutex::Autolock l(mEventLock);
            waitForCommandComplete(OMX_CommandFlush, port);
        }
    }
}

void
OMXComponentDecodeHelper::process_event(void){
    OMX_PARAM_PORTDEFINITIONTYPE port_def;

    // protect mEvents
    Mutex::Autolock l(mEventLock);
    if(!mEvents.empty()) {
	    List<OMXEvent>::iterator it = mEvents.begin();
        switch (it->event){
            ALOGV("%s event: %s",
                __FUNCTION__,
                it->describe());

            case OMX_EventPortSettingsChanged:

                ALOGD("[realtek] process OMX_EventPortSettingsChanged out_port_def.nBufferCountActual %d",
                    out_port_def.nBufferCountActual);

                disablePort(outputPort());
                // wait all buffers returned
                while(true){
#if 0
                    ALOGD("[%s:%d] mOutputBuffers %d",
                        __FUNCTION__,
                        __LINE__,
                        mOutputBuffers.size());
#endif

                    if(mOutputBuffers.size() == out_port_def.nBufferCountActual){
                        printf("[realtek] waited all buffer ######\n");
                        break;
                    }
                    {
                        Mutex::Autolock l(mOutBufferLock);
                        mOutputBufferCondition.waitRelative(mOutBufferLock,5*1E9);
                    }
                }

                cleanupOutputBuffer();
                waitForCommandComplete(OMX_CommandPortDisable, outputPort(), 5*1E9);

                OMX_PARAM_PORTDEFINITIONTYPE port_def;
                get_omx_port_definition(outputPort(), &port_def);
                update_omx_port_definition(outputPort(), &port_def);
#ifdef HAL3_HWC_BUFFER
#if (USE_BUFFER)
                struct EnableAndroidNativeBuffersParams param;
                param.nPortIndex = mOutput;
                param.enable = OMX_TRUE;

                OMX_SetParameter(
                        handle(),
                        (OMX_INDEXTYPE) OMX.google.android.index.enableAndroidNativeBuffers,
                        &param);
#endif
#endif
                enablePort(outputPort());
#if (USE_BUFFER)
                initOutputBuffers2();
#else
                initOutputBuffers();
#endif
                waitForCommandComplete(OMX_CommandPortEnable, outputPort(), 5*1E9);

                ALOGD("OMX_EventPortSettingsChanged enable output port complete [%d]",__LINE__);

                for(int i = 0; i < out_port_def.nBufferCountActual; i++){
                    OMX_BUFFERHEADERTYPE *out_buffer = outputBuffer();
                    fillBuffer(out_buffer);
                }

                mEvents.erase(it);
                ALOGD("OMX_EventPortSettingsChanged complete [%d]",__LINE__);
                break;
            case OMX_EventError:
            case OMX_EventMark:
            case OMX_EventBufferFlag:
                ALOGD("[realtek] process event %d", it->event);
                mEvents.erase(it);
                break;
            default:
                ALOGW("process_event default %d cmd %d",
                    it->event,
                    it->data1);

                mEvents.erase(it);
                // TODO
                mEvents.push_back(*it);
                break;
        }
    }
}

void
OMXComponentDecodeHelper::process_decode(void){
    bool empty = false;

    {
	    Mutex::Autolock l(mOutBufferLock);
        empty = mOutputBuffers.empty();
    }

	if(!empty){
        OMX_BUFFERHEADERTYPE *out_buffer = outputBuffer();
        fillBuffer(out_buffer);
	}
}

// --------------- InputBufferThread -------------- //
OMXComponentDecodeHelper::InputBufferThread::InputBufferThread(
    OMXComponentDecodeHelper *parent)
{
    ALOGD("InputBufferThread()");
    mParent = parent;
    mRunning = true;
}

OMXComponentDecodeHelper::InputBufferThread::~InputBufferThread()
{
    ALOGD("~InputBufferThread()");
}

bool
OMXComponentDecodeHelper::InputBufferThread::threadLoop()
{
    // get a input buffer
    if(mRunning) {

        OMX_BUFFERHEADERTYPE *buffer = mParent->inputBuffer();

        while(buffer == NULL && mRunning) {
            {
                Mutex::Autolock l(mParent->mInBufferLock);
                mParent->mInputBufferCondition.waitRelative(mParent->mInBufferLock,1E9);
            }
            buffer = mParent->inputBuffer();
            ALOGD("InputBufferThread - wait input buffer[%d]",
                __LINE__);
        }

        if(buffer == NULL) {
            ALOGD("InputBufferThread - no input buffer");
            return true;
        }

        nsecs_t pts = -1;
        size_t range = -1;

        CHECK(mParent->mCallback != NULL);

        status_t rst = mParent->mCallback->onInputBufferAvailable(
            buffer->pBuffer,
            0x0,    // no phyAddr
            buffer->nAllocLen,
            &pts,
            &range);

        if(pts != -1 && range != -1) {
            //CHECK(pts != -1);
            //CHECK(range != -1);
            buffer->nFilledLen = range;
            buffer->nTimeStamp = pts;
            buffer->nFlags = OMX_BUFFERFLAG_ENDOFFRAME;
            mParent->emptyBuffer(buffer);
        }else{
            ALOGD("No valid input frame, push back to input queue");
            Mutex::Autolock l(mParent->mInBufferLock);
            mParent->mInputBuffers.push_back(buffer);
        }

        if(rst == INVALID_OPERATION) {
            return false;
        }

        return true;
    }else{
        return false;
    }
}

// --------------- EventThread -------------- //
OMXComponentDecodeHelper::EventThread::EventThread(
    OMXComponentDecodeHelper *parent)
{
    ALOGD("EventThread()");
    mParent = parent;
    mRunning = true;
}

OMXComponentDecodeHelper::EventThread::~EventThread()
{
    ALOGD("~EventThread()");
}

bool
OMXComponentDecodeHelper::EventThread::threadLoop()
{
    if(mRunning) {
		usleep(10000);
        if(mParent->mExecuting) {
            mParent->process_decode();
        }
        mParent->process_event();
        return true;
    }else{
        return false;
    }
}

};  // namespace androidcamera3
